#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
#  VeriTAR: In-place verification of the MD5 sums of files within a tar archive.
#
#  Project: https://www.codetrax.org/projects/veritar
#
#  Copyright 2007 George Notaras <gnot [at] g-loaded.eu>, CodeTRAX.org
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

__version__ = "0.2.0"
__author__ = "George Notaras <gnot -at- g-loaded.eu>"
__credits__ = ""

import os
import sys
import tarfile
#import hashlib
import md5	# for compatibility with older Python versions
from optparse import OptionParser
import time


def err(msg, opts):
	"""Writes 'msg' to stderr.
	"""
	if not opts.quiet:
		sys.stderr.write("%s\n" % msg)
		sys.stderr.flush()

def warn(msg, opts):
	"""Writes 'msg' to stderr.
	"""
	if not opts.quiet:
		if not opts.nowarn:
			sys.stderr.write("%s\n" % msg)
			sys.stderr.flush()

def get_member_md5sum(f):
	"""Returns the archived file's md5 sum.
	
	Accepts the file descriptor of the tar member.
	"""
	READ_BLOCK = 65536
	#m = hashlib.md5()
	m = md5.new()
	data = f.read(READ_BLOCK)
	while data:
		m.update(data)
		data = f.read(READ_BLOCK)
	return m.hexdigest()

def get_valid_checksums(path):
	"""Returns a dictionary of items: 'name:md5sum'
	
	Reads path and retrieves md5 sums and pathnames, which are stored in a
	dictionary using the format:   'name' : md5sum
	"""
	csums = {}
	f_sums = open(path)
	for line in f_sums:
		if line.strip():
			csum, name = line.split("  ")
			csums[name.strip()] = csum.strip()
	f_sums.close()
	return csums

def check_tar_members(tarfile_path, md5file_path, opts):
	"""Performs the md5sum comparison.
	"""
	MSG_OK = "OK"
	MSG_NOOK = "CORRUPT"
	MSG_MISS = "MISSING"
	L_JUSTIFY = 10
	#IND_POS = 0	# Initial position of the progress indicator
	#IND_STATES = ["|", "/", "-", "\\", "*"]
	counters = [0, 0, 0, 0, 0]	# Processed, OK, skipped, NOOK, Missing
	csums = get_valid_checksums(md5file_path)
	f_tar = tarfile.open(tarfile_path, "r|*")
	f_tar.ignore_zeros = True
	f_tar.debug = 1
	f_tar.errorlevel = 2
	sys.stdout.write("%s v%s\n\n" % \
				(os.path.basename(sys.argv[0]), __version__))
	sys.stdout.flush()
	member = f_tar.next()
	if member:
		counters[0] += 1
	while member:
		if member.isfile():
			f = f_tar.extractfile(member)
			checksum = get_member_md5sum(f)
			f.close()
			if checksum == csums[member.name]:
				# member is verified
				if opts.verbose:
					sys.stdout.write("%s %s\n" % \
					(MSG_OK.ljust(L_JUSTIFY), member.name))
					sys.stdout.flush()
				#else:
				#	sys.stdout.write("\b%s" % \
				#			IND_STATES[IND_POS])
				#	sys.stdout.flush()
				#	IND_POS += 1
				#	if IND_POS > len(IND_STATES)-1:
				#		IND_POS = 0
					
				counters[1] += 1	# OK
			else:
				# member is corrupted
				err("%s %s" % \
				(MSG_NOOK.ljust(L_JUSTIFY), member.name), opts)
				counters[3] += 1	# FAILED
			if csums.has_key(member.name):
				del csums[member.name]
			else:
				# there is no checksum for member
				# in the checksum file (wrong checksum file)
				# It does not mean a damaged archive (TODO)
				warn("%s: no checksum for member: %s" % \
				("WARNING".ljust(L_JUSTIFY), member.name), opts)
							# NO COUNT
		else:
			if csums.has_key(member.name):
				# member is not a file, but a checksum exists
				# in the checksum file
				del csums[member.name]
			warn("%s SKIPPING: %s (%s)" % \
					("WARNING".ljust(L_JUSTIFY), \
					member.name, member.type), opts)
			counters[2] += 1	# SKIP
		member = f_tar.next()
		if member:
			counters[0] += 1
	f_tar.close()
	if csums:
		# If the checksums-file contains more items than the TAR members
		# it is assumed that the archive is corrupted.
		for remnant in csums.keys():
			err("%s %s" % \
				(MSG_MISS.ljust(L_JUSTIFY), remnant), opts)
			counters[4] += 1	# MISSING
	summary(counters)

def summary(counters):
	# Processed, OK, skipped, NOOK, Missing
	L_JUST = 10
	title = "TAR members checksum verification"
	sys.stdout.write("\n%s\n" % title)
	sys.stdout.write("%s\n" % ("-" * len(title)))
	sys.stdout.write("%s: %s\n" % ("Processed".ljust(L_JUST), counters[0]))
	sys.stdout.write("%s: %s\n" % ("Verified".ljust(L_JUST), counters[1]))
	sys.stdout.write("%s: %s\n" % ("Skipped".ljust(L_JUST), counters[2]))
	sys.stdout.write("%s: %s\n" % ("Failed".ljust(L_JUST), counters[3]))
	sys.stdout.write("%s: %s\n" % ("Missing".ljust(L_JUST), counters[4]))
	sys.stdout.write("%s\n" % ("-" * len(title)))
	if counters[3] or counters[4]:
		sys.stdout.write("FAILED\n")
	else:
		sys.stdout.write("SUCCESS\n")
	sys.stdout.write("%s\n" % ("-" * len(title)))
	sys.stdout.flush()

def parse_cli():
	usage = """

    %prog [options] tar_archive checksum_file

    md5sumfile format:  'md5sum  path'"""
	parser = OptionParser(usage=usage, version=__version__)
	parser.add_option("-v", "--verbose", action="store_true", \
		dest="verbose", help= \
		"""Print all messages. Cannot be used with -q.""")
	parser.add_option("-q", "--quiet", action="store_true", dest="quiet",
		help="""Only checksum errors will be printed. Warnings are \
suppressed. Cannot be used with -v.""")
	parser.add_option("-n", "--no-warn", action="store_true", dest="nowarn",
		help="""Warnings are suppressed. Note that using this switch \
together with -q has absolutely no effect, since -q suppresses warnings \
anyway.""")
	opts, args = parser.parse_args()
	if len(args) != 2:
		parser.error("Wrong number of arguments")
	elif opts.quiet and opts.verbose:
		parser.error("Cannot use -v and -q together.")
	elif not os.path.exists(args[0]) or not os.path.exists(args[1]):
		parser.error("Invalid file path")
	return opts, args

def main():
	start = time.time()
	opts, args = parse_cli()
	tar_path, csum_path = args
	try:
		check_tar_members(tar_path, csum_path, opts)
	except KeyboardInterrupt:
		err("Operation aborted by user", opts)
		sys.exit(1)
	else:
		finish = time.time()
		sys.stdout.write("Elapsed: %.3f sec\n" % (finish -start))
		sys.stdout.flush()

if __name__ == '__main__':
	main()

